SETLOCAL
SET ARGS=%*

ECHO "THIS SCRIPT IS DEPRECATED - USE PY27_MSVC_BUILD.BAT INSTEAD"
REM support double-click on BAT file, so run from parent directory:
IF EXIST "MAKE-INSTALLER.BAT" CD ..

win32\PY27_MSVC_BUILD.BAT %ARGS%

ENDLOCAL
